library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
library SPI;
use SPI.core_if.ALL;

-- ------------------------------------------------------------------
-- Module:  SPICore
-- Purpose: perform serial communication using SPI-Protocol
--          This core supports both byte orders (msb-first and lsb-first),
--          as well as all 4 SPI modes and asynchronous reset.
--          o_core.byte_done is feedback to processor of parallel data.
-- Author:  myself
-- Created: 20090809
-- ------------------------------------------------------------------
ENTITY SPICore IS
    Port     ( in_clock  : in  std_logic;         -- spi clock from client
                                                      
               reset     : in  std_logic;         -- asynchronous reset                               
               i_core    : in  SPI_core_in_type;  -- input from SPI-client/busmaster
               o_core    : out SPI_core_out_type; -- output to SPI-client/busmaster

               -- SPI pins of bus lines (slave select is driven by client/busmaster)
               spi_clock : out std_logic;         -- SPI line: clock
               spi_miso  : in  std_logic;         -- SPI line: incoming data
               spi_mosi  : out std_logic          -- SPI line: outgoing data
             );
END SPICore;

ARCHITECTURE std of SPICore IS
   type reg_type is record 
      i_core  : SPI_core_in_type;
      o_core  : SPI_core_out_type;
      bcnt    : integer range 0 to 7;             -- used to count the transmitted bits
      oData   : std_logic_vector(7 downto 0);     -- buffer for outgoing data
      iData   : std_logic_vector(7 downto 0);     -- buffer for incoming data
      miso    : std_logic;
      mosi    : std_logic;
   end record;
   constant rdef : reg_type := (
      i_core => ( dwrite    => (others=>'0'),
                  msb_first => '1',
                  mode      => (others=>'0')),
      o_core => ( dread     => (others=>'0'),
                  byte_done => '0' ),
      bcnt   => 0,
      oData  => (others=>'0'),
      iData  => (others=>'0'),
      miso   => '0',
      mosi   => '0'
   );
   
   -- ---------------------------------------------------------------
   -- shift serial data depending SPI-mode and byte order
   -- ---------------------------------------------------------------
   FUNCTION processStep(r    : reg_type;
                        clk  : std_logic;
                        miso : std_logic) RETURN reg_type IS
      variable tmp : reg_type;
   BEGIN
      tmp := r;

      if (r.bcnt < 7) then
         if (r.bcnt = 0) then
            tmp.iData := r.i_core.dwrite;
            tmp.oData := r.o_core.dread;
         elsif (r.bcnt = 4) then
            tmp.o_core.byte_done := '0';         -- 4 serial ticks time to process parallel data
         end if;   

         case tmp.i_core.mode is
         -- sample (rising), setup (falling)
         when "00" | "11" => if (clk = '0') then
                                if (r.i_core.msb_first = '1') then
                                   tmp.iData := tmp.iData(7 downto 1) & miso; 
                                else
                                   tmp.iData := miso & tmp.iData(6 downto 0);
                                end if;
                             else
                                if (r.i_core.msb_first = '1') then
                                   tmp.mosi  := tmp.oData(7);
                                   tmp.oData := tmp.oData(6 downto 0) & '0';
                                else
                                   tmp.mosi  := tmp.oData(0);
                                   tmp.oData := '0' & tmp.oData(7 downto 1);
                                end if;
                             end if;
         -- setup (rising), sample (falling)
         when "01" | "10" => if (clk = '0') then
                                if (r.i_core.msb_first = '1') then
                                   tmp.mosi  := tmp.oData(7);
                                   tmp.oData := tmp.oData(6 downto 0) & '0';
                                else
                                   tmp.mosi  := tmp.oData(0);
                                   tmp.oData := '0' & tmp.oData(7 downto 1);
                                end if;
                             else
                                if (r.i_core.msb_first = '1') then
                                   tmp.iData := tmp.iData(7 downto 1) & miso; 
                                else
                                   tmp.iData := miso & tmp.iData(6 downto 0);
                                end if;
                             end if;
         when others => NULL;
         end case;
         tmp.bcnt := tmp.bcnt + 1;
      else
         tmp.bcnt := 0;
         tmp.o_core.dread := tmp.iData;
         tmp.o_core.byte_done := '1';
      end if;

      return tmp;
   END;
   signal r, rin : reg_type;
BEGIN
   spi_clock <= in_clock;


   comp: PROCESS(r, in_clock, i_core, spi_miso)
      variable v : reg_type;
   BEGIN
      v := r;           
      v.i_core := i_core;
      spi_mosi <= r.mosi;
      o_core   <= r.o_core;
      rin <= processStep(v, in_clock, spi_miso);
   END PROCESS;

   
   regs : PROCESS(in_clock, reset)
   BEGIN
      if    reset = '1'           then r <= rdef;
      elsif rising_edge(in_clock) then r <= rin; 
      end if;
   END PROCESS;
END std;
